"""Identify mnist digits."""
import argparse
import struct
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
from flax import linen as nn
from tqdm import tqdm

from input_opt import CNN


def get_mnist_test_data() -> Tuple[np.ndarray, np.ndarray]:
    """Return the mnist test data set in numpy arrays.

    Returns:
        (array, array): A touple containing the test
        images and labels.
    """
    with open("./data/MNIST/t10k-images-idx3-ubyte", "rb") as f:
        _, size = struct.unpack(">II", f.read(8))
        nrows, ncols = struct.unpack(">II", f.read(8))
        data = np.array(np.fromfile(f, dtype=np.dtype(np.uint8).newbyteorder(">")))
        img_data_test = data.reshape((size, nrows, ncols))

    with open("./data/MNIST/t10k-labels-idx1-ubyte", "rb") as f:
        _, size = struct.unpack(">II", f.read(8))
        lbl_data_test = np.array(np.fromfile(f, dtype=np.dtype(np.uint8)))
    # if gpu:
    #    return cp.array(img_data_test), cp.array(lbl_data_test)
    return img_data_test, lbl_data_test


def get_mnist_train_data() -> Tuple[np.ndarray, np.ndarray]:
    """Load the mnist training data set.

    Returns:
        (array, array): A touple containing the training
        images and labels.
    """
    with open("./data/MNIST/train-images-idx3-ubyte", "rb") as f:
        _, size = struct.unpack(">II", f.read(8))
        nrows, ncols = struct.unpack(">II", f.read(8))
        data = np.array(np.fromfile(f, dtype=np.dtype(np.uint8).newbyteorder(">")))
        img_data_train = data.reshape((size, nrows, ncols))

    with open("./data/MNIST/train-labels-idx1-ubyte", "rb") as f:
        _, size = struct.unpack(">II", f.read(8))
        lbl_data_train = np.array(np.fromfile(f, dtype=np.dtype(np.uint8)))
    # if gpu:
    #    return cp.array(img_data_train), cp.array(lbl_data_train)
    return img_data_train, lbl_data_train


@jax.jit
def normalize(
    data: np.ndarray, mean: Optional[float] = None, std: Optional[float] = None
) -> Tuple[np.ndarray, float, float]:
    """Normalize the input array.

    After normalization the input
    distribution should be approximately standard normal.

    Args:
        data (np.array): The input array.
        mean (float): Data mean, re-computed if None.
            Defaults to None.
        std (float): Data standard deviation,
            re-computed if None. Defaults to None.

    Returns:
        np.array, float, float: Normalized data, mean and std.
    """
    if mean is None:
        mean = np.mean(data)
    if std is None:
        std = np.std(data)
    return (data - mean) / std, mean, std


@jax.jit
def cross_entropy(label: jnp.ndarray, out: jnp.ndarray) -> jnp.ndarray:
    """Compute the cross entropy of one-hot encoded labels and the network output.

    Args:
        label (jnp.ndarray): The image labels of shape [batch_size, 10].
        out (jnp.ndarray): The network output of shape [batch_size, 10].

    Returns:
        jnp.ndarray, The loss scalar.

    """
    left = -label * jnp.log(out + 1e-8)
    right = -(1 - label) * jnp.log(1 - out + 1e-8)
    return jnp.mean(left + right)


@jax.jit
def forward_pass(weights, img_batch, label_batch):
    """Do a forward step."""
    out = cnn.apply(weights, jnp.expand_dims(img_batch, -1))
    ce_loss = cross_entropy(nn.one_hot(label_batch, num_classes=10), out)
    return ce_loss


# set up autograd
loss_grad_fn = jax.value_and_grad(forward_pass)


# set up SGD
@jax.jit
def sgd_step(weights, grads, learning_rate):
    """Update the variable in a SGD step."""
    weights = jax.tree_util.tree_map(lambda p, g: p - learning_rate * g, weights, grads)
    return weights


def get_acc(img_data, label_data):
    """Compute the network accuracy."""
    out = cnn.apply(weights, jnp.expand_dims(img_data, -1))
    rec = jnp.argmax(out, axis=1)
    acc = jnp.sum((rec == label_data).astype(np.float32)) / len(label_data)
    return acc


def integrate_gradients(net, weights, test_images, output_digit, steps_m=300):
    """Calculate integrated gradients."""
    g_list = []
    for test_image_x in tqdm(test_images, desc="Integrating Gradients"):
        # TODO: create a list for the gradients.
        # step_g_list = []
        # TODO: create a black reference image using jnp.zeros_like .
        # black_image_x_prime = jnp.zeros_like(test_image_x)
        # TODO: Loop over the integration steps.
        # for current_step_k in range(steps_m):
        #  # TODO: compute the input to F from equation 5 in the slides.
        #  current_point = black_image_x_prime + (current_step_k/steps_m) * (test_image_x - black_image_x_prime)
        #  # TODO: define a forward pass for jax.grad
        #  def eval_fun(point):
        #    logits = net.apply(weights, point)
        #    return logits[0, output_digit]
        #  # TODO: use jax.grad to find the gradient with repsect to the input image.
        #  grad_fun = jaxday_14_exercise_interpretability_solution/scripts/train.slurm.grad(eval_fun)
        #  grads = grad_fun(jnp.expand_dims(current_point, 0))
        #  # TODO: append the gradient to yout list
        #  step_g_list.append(grads)
        # TODO: Return the sum of the of the list elements.
        # g_sum = jnp.sum(jnp.concatenate(step_g_list), 0)

        test_image_x = jnp.expand_dims(jnp.expand_dims(test_image_x, -1), 0)
        black_image_x_prime = jnp.zeros_like(test_image_x)
        scales = jnp.expand_dims(jnp.arange(steps_m) / steps_m, (1, 2, 3))
        points = black_image_x_prime + scales * (test_image_x - black_image_x_prime)

        def eval_fun(point):
            logits = net.apply(weights, point)
            return jnp.sum(logits[:, output_digit], 0)

        grad_fun = jax.grad(eval_fun)
        grads = grad_fun(points)
        g_sum = jnp.sum(grads, 0)
        integrated_g = (test_image_x - black_image_x_prime) * g_sum * (1.0 / steps_m)
        g_list.append(integrated_g)
    return jnp.mean(jnp.concatenate(g_list, 0), 0)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Train Networks on MNIST.")
    parser.add_argument("--lr", type=float, default=0.0025, help="Learning Rate")
    args = parser.parse_args()
    print(args)

    batch_size = 50
    val_size = 500
    epochs = 20
    img_data_train, lbl_data_train = get_mnist_train_data()
    img_data_val, lbl_data_val = img_data_train[:val_size], lbl_data_train[:val_size]
    img_data_train, lbl_data_train = (
        img_data_train[val_size:],
        lbl_data_train[val_size:],
    )
    img_data_train, mean, std = normalize(img_data_train)
    img_data_val, _, _ = normalize(img_data_val, mean, std)

    key = jax.random.PRNGKey(42)  # type: ignore
    cnn = CNN()
    weights = cnn.init(
        key, jnp.ones([batch_size] + list(img_data_train.shape[1:]) + [1])
    )

    for e in range(epochs):
        shuffler = jax.random.permutation(key, len(img_data_train))
        img_data_train = img_data_train[shuffler]
        lbl_data_train = lbl_data_train[shuffler]

        img_batches = np.split(
            img_data_train, img_data_train.shape[0] // batch_size, axis=0
        )
        label_batches = np.split(
            lbl_data_train, lbl_data_train.shape[0] // batch_size, axis=0
        )

        bar = tqdm(
            zip(img_batches, label_batches),
            total=len(img_batches),
            desc="Training CNN",
        )

        for img_batch, label_batch in bar:
            img_batch, label_batch = (
                jnp.array(np_array) for np_array in (img_batch, label_batch)
            )
            # cel = cross_entropy(nn.one_hot(label_batches[no], num_classes=10),
            #                    out)
            cel, grads = loss_grad_fn(weights, img_batch, label_batch)
            weights = sgd_step(weights, grads, args.lr)
            bar.set_description("Loss: {:2.3f}".format(cel))
        print("Epoch: {}, loss: {}".format(e, cel))

        # train_acc = get_acc(img_data_train, lbl_data_train)
        val_acc = get_acc(img_data_val, lbl_data_val)
        print("Validation accuracy: {:3.3f}".format(val_acc))

    print("Training done. Testing...")
    img_data_test, lbl_data_test = get_mnist_test_data()
    img_data_test, mean, std = normalize(img_data_test, mean, std)

    test_acc = get_acc(img_data_test, lbl_data_test)
    print("Done. Test acc: {}".format(test_acc))

    if 1:
        print("IG for digit 0")
        ig_0 = integrate_gradients(
            cnn, weights, img_data_test[:500], output_digit=0, steps_m=300
        )
        print("IG for digit 1")
        plt.imshow(ig_0)
        plt.savefig("integrated_gradients_0.jpg")
        ig_1 = integrate_gradients(
            cnn, weights, img_data_test[:500], output_digit=1, steps_m=300
        )
        plt.imshow(ig_1)
        plt.savefig("integrated_gradients_1.jpg")
        print("stop")
